from flask import Flask
import unittest
from flask.ext.testing import TestCase as FlaskTestCase
from flask.ext.sqlalchemy import SQLAlchemy
from .core import Permissions
from .utils import is_sequence
from .decorators import user_has, user_is
from werkzeug.exceptions import Forbidden
import os

app = Flask(__name__)
app.config['TESTING'] = True

db_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'app.db')

app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///' + db_path

db = SQLAlchemy(app)

perms = Permissions(app, db, None)

from .models import Role, Ability, UserMixin


class DatabaseTests(FlaskTestCase):

    def create_app(self):
        return app

    def setUp(self):
        db.create_all()

    def tearDown(self):
        os.remove(db_path)


class ModelsTests(DatabaseTests):

    def test_user_mixin(self):
        user = UserMixin()
        db.session.add(user)
        db.session.commit()
        self.assertEqual(user.id, 1)

    def test_new_user_with_roles_assigned(self):
        roles = ['admin', 'superadmin']
        for role in roles:
            new_role = Role(role)
            db.session.add(new_role)
            db.session.commit()
        user = UserMixin(roles=roles, default_role=None)
        db.session.add(user)
        db.session.commit()
        self.assertEqual(user.id, 1)

    def test_new_user_with_a_single_role_assigned(self):
        new_role = Role('admin')
        db.session.add(new_role)
        db.session.commit()
        user = UserMixin(roles='admin', default_role=None)
        db.session.add(user)
        db.session.commit()
        self.assertEqual(user.id, 1)

    def test_user_is_authenticated(self):
        user = UserMixin()
        self.assertTrue(user.is_authenticated())

    def test_user_is_active(self):
        user = UserMixin()
        self.assertTrue(user.is_active())

    def test_user_is_not_anonymous(self):
        user = UserMixin()
        self.assertFalse(user.is_anonymous())

    def test_user_get_id(self):
        user = UserMixin()
        db.session.add(user)
        db.session.commit()
        self.assertEqual(user.get_id(), unicode(1))

    def test_user_repr(self):
        user = UserMixin()
        db.session.add(user)
        db.session.commit()
        self.assertEqual(user.__repr__(), '<User 1>')

    def test_role(self):
        role = Role('admin')
        db.session.add(role)
        db.session.commit()
        self.assertEqual(role.id, 1)

    def test_role_repr(self):
        role = Role('admin')
        self.assertEqual(role.__repr__(), '<Role admin>')

    def test_role_str(self):
        role = Role('admin')
        self.assertEqual(role.__str__(), 'admin')

    def test_ability(self):
        ability = Ability('create_users')
        db.session.add(ability)
        db.session.commit()
        self.assertEqual(ability.id, 1)

    def test_ability_repr(self):
        ability = Ability('create_users')
        self.assertEqual(ability.__repr__(), '<Ability create_users>')

    def test_ability_repr(self):
        ability = Ability('create_users')
        self.assertEqual(ability.__str__(), 'create_users')

    def test_add_roles_with_nonexisting_roles(self):
        user = UserMixin(default_role=None)
        new_roles = ['admin', 'superadmin', 'editor']
        user.add_roles(*new_roles)
        db.session.add(user)
        db.session.commit()
        test_user = UserMixin.query.get(1)
        self.assertItemsEqual(new_roles, test_user.roles)

    def test_add_roles_with_existing_roles(self):
        user = UserMixin(default_role=None)
        new_roles = ['admin', 'superadmin', 'editor']
        for role in new_roles:
            new_role = Role(role)
            db.session.add(new_role)
            db.session.commit()
        user.add_roles(*new_roles)
        db.session.add(user)
        db.session.commit()
        test_user = UserMixin.query.get(1)
        self.assertItemsEqual(new_roles, test_user.roles)

    def test_remove_roles(self):
        user = UserMixin()
        new_roles = ['admin', 'user', 'superadmin']
        user.add_roles(*new_roles)
        db.session.add(user)
        db.session.commit()
        role_remove_user = UserMixin.query.get(1)
        role_remove_user.remove_roles('user', 'admin')
        db.session.add(user)
        db.session.commit()
        test_user = UserMixin.query.get(1)
        self.assertItemsEqual(test_user.roles, ['superadmin'])

    def test_add_abilities_with_nonexisting_abilities(self):
        role = Role('admin')
        new_abilities = ['create_users', 'set_roles', 'set_abilities']
        role.add_abilities(*new_abilities)
        db.session.add(role)
        db.session.commit()
        test_role = Role.query.get(1)
        abilities = [ability.name for ability in test_role.abilities]
        self.assertItemsEqual(new_abilities, abilities)

    def test_add_abilities_with_existing_abilities(self):
        role = Role('admin')
        new_abilities = ['create_users', 'set_roles', 'set_abilities']
        for ability in new_abilities:
            new_ability = Ability(ability)
            db.session.add(new_ability)
            db.session.commit()
        role.add_abilities(*new_abilities)
        db.session.add(role)
        db.session.commit()
        test_role = Role.query.get(1)
        abilities = [ability.name for ability in test_role.abilities]
        self.assertItemsEqual(new_abilities, abilities)

    def test_remove_abilities(self):
        role = Role('admin')
        new_abilities = ['create_users', 'set_roles', 'set_abilities']
        role.add_abilities(*new_abilities)
        db.session.add(role)
        db.session.commit()
        ability_remove_role = Role.query.get(1)
        ability_remove_role.remove_abilities('set_roles', 'set_abilities')
        db.session.add(ability_remove_role)
        db.session.commit()
        test_role = Role.query.get(1)
        abilities = [ability.name for ability in test_role.abilities]
        self.assertItemsEqual(abilities, ['create_users'])


class DecoratorsTests(DatabaseTests):

    def mock_function(self):
        return True

    def create_user(self):
        role = Role('admin')
        new_abilities = ['create_users', 'set_roles', 'set_abilities']
        for ability in new_abilities:
            new_ability = Ability(ability)
            db.session.add(new_ability)
            db.session.commit()
        role.add_abilities(*new_abilities)
        db.session.add(role)
        db.session.commit()

        user = UserMixin(roles='admin')
        db.session.add(user)
        db.session.commit()

    def return_user(self):
        user = UserMixin.query.get(1)
        return user

    def setUp(self):
        super(DecoratorsTests, self).setUp()
        self.create_user()

    def test_user_has_pass(self):
        wrapped_function = user_has(
            'create_users', self.return_user)(self.mock_function)
        self.assertTrue(wrapped_function())

    def test_user_has_fail(self):
        wrapped_function = user_has(
            'edit', self.return_user)(self.mock_function)
        self.assertRaises(Forbidden, wrapped_function)

    def test_user_is_pass(self):
        wrapped_function = user_is(
            'admin', self.return_user)(self.mock_function)
        self.assertTrue(wrapped_function())

    def test_user_is_fail(self):
        wrapped_function = user_is(
            'user', self.return_user)(self.mock_function)
        self.assertRaises(Forbidden, wrapped_function)


class UtilsTests(unittest.TestCase):

    def test_is_string_a_sequence(self):
        string_var = 'This is a string. It is not a sequence.'
        self.assertFalse(is_sequence(string_var))

    def test_is_integer_a_sequence(self):
        int_var = 1
        self.assertFalse(is_sequence(int_var))

    def test_is_bool_a_sequence(self):
        bool_var = True
        self.assertFalse(is_sequence(bool_var))

    def test_is_list_a_sequence(self):
        list_var = [1, 2, 3]
        self.assertTrue(is_sequence(list_var))

    def test_is_set_a_sequence(self):
        set_var = set([1, 2, 3])
        self.assertTrue(is_sequence(set_var))

    def test_is_tuple_a_sequence(self):
        tuple_var = (1, 2, 3)
        self.assertTrue(is_sequence(tuple_var))

    def test_is_dict_a_sequence(self):
        dict_var = {'a': 1, 'b': 2, 'c': 3}
        self.assertTrue(is_sequence(dict_var))



try:
    from .core import db
except ImportError:
    raise Exception(
        'Permissions app must be initialized before importing models')

from werkzeug import generate_password_hash, check_password_hash
from sqlalchemy.ext.associationproxy import association_proxy
from .utils import is_sequence


def _role_find_or_create(r):
    role = Role.query.filter_by(name=r).first()
    if not(role):
        role = Role(name=r)
        db.session.add(role)
    return role


user_role_table = db.Table('fp_user_role',
                           db.Column(
                               'user_id', db.Integer, db.ForeignKey('fp_user.id')),
                           db.Column(
                           'role_id', db.Integer, db.ForeignKey('fp_role.id'))
                           )

role_ability_table = db.Table('fp_role_ability',
                              db.Column(
                                  'role_id', db.Integer, db.ForeignKey('fp_role.id')),
                              db.Column(
                              'ability_id', db.Integer, db.ForeignKey('fp_ability.id'))
                              )


class Role(db.Model):

    """
    Subclass this for your roles
    """
    __tablename__ = 'fp_role'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(120), unique=True)
    abilities = db.relationship(
        'Ability', secondary=role_ability_table, backref='roles')

    def __init__(self, name):
        print name
        print type(name)
        self.name = name.lower()

    def add_abilities(self, *abilities):
        for ability in abilities:
            existing_ability = Ability.query.filter_by(
                name=ability).first()
            if not existing_ability:
                existing_ability = Ability(ability)
                db.session.add(existing_ability)
                db.session.commit()
            self.abilities.append(existing_ability)

    def remove_abilities(self, *abilities):
        for ability in abilities:
            existing_ability = Ability.query.filter_by(name=ability).first()
            if existing_ability and existing_ability in self.abilities:
                self.abilities.remove(existing_ability)

    def __repr__(self):
        return '<Role {}>'.format(self.name)

    def __str__(self):
        return self.name


class Ability(db.Model):

    """
    Subclass this for your abilities
    """
    __tablename__ = 'fp_ability'
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(120), unique=True)

    def __init__(self, name):
        self.name = name.lower()

    def __repr__(self):
        return '<Ability {}>'.format(self.name)

    def __str__(self):
        return self.name


class UserMixin(db.Model):

    """
    Subclass this for your user class
    """
    __tablename__ = 'fp_user'
    id = db.Column(db.Integer, primary_key=True)
    _roles = db.relationship(
        'Role', secondary=user_role_table, backref='users')
    type = db.Column(db.String(50))

    roles = association_proxy('_roles', 'name', creator=_role_find_or_create)

    __mapper_args__ = {
        'polymorphic_identity': 'usermixin',
        'polymorphic_on': type
    }

    def __init__(self, roles=None, default_role='user'):
        # If only a string is passed for roles, convert it to a list containing
        # that string
        if roles and isinstance(roles, basestring):
            roles = [roles]

        # If a sequence is passed for roles (or if roles has been converted to
        # a sequence), fetch the corresponding database objects and make a list
        # of those.
        if roles and is_sequence(roles):
            self.roles = roles
        # Otherwise, assign the default 'user' role. Create that role if it
        # doesn't exist.
        elif default_role:
            self.roles = [default_role]

    def is_authenticated(self):
        return True

    def is_active(self):
        return True

    def is_anonymous(self):
        return False

    def add_roles(self, *roles):
        self.roles.extend([role for role in roles if role not in self.roles])

    def remove_roles(self, *roles):
        self.roles = [role for role in self.roles if role not in roles]

    def get_id(self):
        return unicode(self.id)

    def __repr__(self):
        return '<User {}>'.format(self.id)



class Permissions(object):

    def __init__(self, app, local_db, current_user):
        self.init_app(app, local_db, current_user)

    def init_app(self, app, local_db, current_user):
        self.app = app
        self.current_user = current_user
        global db
        db = local_db



from functools import wraps
from werkzeug.exceptions import Forbidden


def import_user():
    try:
        from flask.ext.login import current_user
        return current_user
    except ImportError:
        raise ImportError(
            'User argument not passed and Flask-Login current_user could not be imported.')


def user_has(ability, get_user=import_user):
    """
    Takes an ability (a string name of either a role or an ability) and returns the function if the user has that ability
    """
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            from .models import Ability
            desired_ability = Ability.query.filter_by(
                name=ability).first()
            user_abilities = []
            current_user = get_user()
            for role in current_user._roles:
                user_abilities += role.abilities
            print "Abilities: {}".format(user_abilities)
            if desired_ability in user_abilities:
                return func(*args, **kwargs)
            else:
                raise Forbidden("You do not have access")
        return inner
    return wrapper


def user_is(role, get_user=import_user):
    """
    Takes an role (a string name of either a role or an ability) and returns the function if the user has that role
    """
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            from .models import Role
            current_user = get_user()
            if role in current_user.roles:
                return func(*args, **kwargs)
            raise Forbidden("You do not have access")
        return inner
    return wrapper
